# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import re
import subprocess
import sys
from pathlib import Path
from typing import List, Optional, Set

import libcst as cst
import libcst.matchers as m

BASE_DIR = (Path.cwd() / Path(__file__)).parent
KNOWN_FILE = BASE_DIR / "known.py"
DOCS_DIR = BASE_DIR.parent / "docs"
API_FILE = DOCS_DIR / "api.rst"

RELEASES = {
    "2.3": "https://www.python.org/ftp/python/2.3.7/Python-2.3.7.tgz",
    "2.4": "https://www.python.org/ftp/python/2.4.6/Python-2.4.6.tgz",
    "2.5": "https://www.python.org/ftp/python/2.5.6/Python-2.5.6.tgz",
    "2.6": "https://www.python.org/ftp/python/2.6.9/Python-2.6.9.tgz",
    "2.7": "https://www.python.org/ftp/python/2.7.18/Python-2.7.18.tgz",
    "3.0": "https://www.python.org/ftp/python/3.0.1/Python-3.0.1.tgz",
    "3.1": "https://www.python.org/ftp/python/3.1.5/Python-3.1.5.tgz",
    "3.2": "https://www.python.org/ftp/python/3.2.6/Python-3.2.6.tgz",
    "3.3": "https://www.python.org/ftp/python/3.3.7/Python-3.3.7.tgz",
    "3.4": "https://www.python.org/ftp/python/3.4.10/Python-3.4.10.tgz",
    "3.5": "https://www.python.org/ftp/python/3.5.10/Python-3.5.10.tgz",
    "3.6": "https://www.python.org/ftp/python/3.6.13/Python-3.6.13.tgz",
    "3.7": "https://www.python.org/ftp/python/3.7.10/Python-3.7.10.tgz",
    "3.8": "https://www.python.org/ftp/python/3.8.8/Python-3.8.8.tgz",
    "3.9": "https://www.python.org/ftp/python/3.9.5/Python-3.9.5.tgz",
    "3.10": "https://www.python.org/ftp/python/3.10.2/Python-3.10.2.tgz",
    "3.11": "https://www.python.org/ftp/python/3.11.0/Python-3.11.0b1.tgz",
}

MODULE_DEF_RE = re.compile(r"PyModuleDef .*? = \{\s*[^,]*,\s*([^,}]+)[,}]")
MULTILINE_COMMENT_RE = re.compile(r"/\*.*?\*/")
PY2_INITMODULE_RE = re.compile(r".Py_InitModule\d?\(\s*(?!\))(.*?),")
INITTAB_SELECTOR_RE = re.compile(
    r"(?:struct _(?:frozen|module_alias) .*?|_PyImport_\w+)\[\] = \{([\w\W]+?)\n\};"
)
INITTAB_RE = re.compile(r'{"([^"]+)", ')

# lib2to3 outputs code that doesn't parse, so just omit these lines
PY2_LINES_TO_OMIT = [
    "join(F, fw + '.framework', H)",
    "for fw in 'Tcl', 'Tk'",
    "for fw in ('Tcl', 'Tk')",
    "for H in 'Headers', 'Versions/Current/PrivateHeaders'",
]


with open(__file__) as f:
    MY_COPYRIGHT_HEADER = ""
    while True:
        _line = f.readline()
        if not _line.strip():
            break
        MY_COPYRIGHT_HEADER += _line

GENERATED_TMPL = (
    MY_COPYRIGHT_HEADER
    + '''\

# Generated by stdlibs/fetch.py

from typing import FrozenSet

module_names: FrozenSet[str] = frozenset(
    [
{lines}    ]
)
"""
Known stdlib modules for Python {version}.
"""
'''
)

KNOWN_TMPL = (
    MY_COPYRIGHT_HEADER
    + '''\

# Generated by stdlibs/fetch.py

from typing import List

KNOWN_VERSIONS: List[str] = [
{lines}]
"""All supported Python major releases"""
'''
)

API_ANCHOR = ".. GENERATED\n"
API_TMPL = """
.. module:: stdlibs.{module}

.. autodata:: module_names
    :no-value:
"""


def write_tmpl(module: str, version: str, data: Set[str]) -> None:
    lines = "".join(f'        "{s}",\n' for s in sorted(data))
    (BASE_DIR / f"{module}.py").write_text(
        GENERATED_TMPL.format(lines=lines, version=version)
    )


def regen_all() -> None:
    all2: Set[str] = set()
    all3: Set[str] = set()
    for v in RELEASES:
        names = regen(v)
        if v.startswith("2"):
            all2 |= names
        elif v.startswith("3"):
            all3 |= names
        else:
            raise ValueError("What is this brave new future you live in")

    write_tmpl("py2", "2.x", all2)
    write_tmpl("py3", "3.x", all3)
    write_tmpl("py", "3.x and 2.x", all2 | all3)

    lines = "".join(f'    "{s}",\n' for s in RELEASES)
    KNOWN_FILE.write_text(KNOWN_TMPL.format(lines=lines))
    print("known.py done")

    api_text = API_FILE.read_text()
    api_text, anchor, _ = api_text.partition(API_ANCHOR)
    api_text += anchor

    api_text += API_TMPL.format(module="py3")
    api_text += API_TMPL.format(module="py2")
    api_text += API_TMPL.format(module="py")

    for v in list(RELEASES)[::-1]:
        module = f"py{v.replace('.','')}"
        api_text += API_TMPL.format(module=module)

    API_FILE.write_text(api_text)
    print("api.rst done")

    print("done")


def regen(version: str) -> Set[str]:
    base_path = Path(".cache", RELEASES[version].split("/")[-1].rsplit(".", 1)[0])
    setup_path = base_path / "setup.py"

    if not base_path.exists():
        Path(".cache").mkdir(exist_ok=True)
        subprocess.check_call(["wget", "-c", RELEASES[version]], cwd=".cache")
        subprocess.check_call(
            ["tar", "-xvzf", RELEASES[version].split("/")[-1]], cwd=".cache"
        )
        if version.startswith("2"):
            (base_path / "fixed").mkdir(exist_ok=True)
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "lib2to3",
                    "-n",
                    "-w",
                    "-o",
                    str(base_path / "fixed"),
                    str(base_path / "setup.py"),
                ]
            )

            # TODO if the extraction succeeded but now this fails, we can end up
            # with a corrupt copy.
            setup_path = base_path / "fixed" / "setup.py"
            lines = setup_path.read_text().splitlines(True)
            lines = [line for line in lines if line.strip() not in PY2_LINES_TO_OMIT]
            setup_path.write_text("".join(lines))
    elif version.startswith("2"):
        setup_path = base_path / "fixed" / "setup.py"

    module = try_parse(setup_path)
    ev = ExtensionVisitor()
    module.visit(ev)

    # Python files
    names = ev.extension_names[:]
    for p in (base_path / "Lib").glob("*"):
        if p.name.startswith(("plat-", "lib-")):
            # 2.x platform dirs, or tk support
            for path in p.iterdir():
                # lib-tk/test on 2.7
                if path.name in ("test",):
                    continue
                # TODO plat-mac/lib-scriptpackages
                if path.is_dir() and not path.name.startswith("lib-"):
                    names.append(path.name)
                elif path.name.endswith(".py"):
                    name = path.with_suffix("").name
                    names.append(name)
        else:
            name = p.with_suffix("").name
            name = name.split(".")[0]  # __phello__.foo
            if name not in ("__pycache__", "site-packages", "test"):
                names.append(name)

    for subdir in (
        "Python",  # builtin
        "Modules",  # other extensions, some of which are built-in :/
        "PC",  # windows
    ):
        for p in (base_path / subdir).glob("*.c"):
            try:
                data = p.read_text()
            except UnicodeDecodeError:
                data = p.read_text(encoding="latin-1")

            match = MODULE_DEF_RE.search(data) or PY2_INITMODULE_RE.search(data)
            if match:
                s = MULTILINE_COMMENT_RE.sub("", match.group(1)).strip()
                if s.startswith(".m_name"):
                    s = s.split("=")[1].strip()

                if s.startswith('"') and s.endswith('"'):
                    names.append(s.strip('"'))
                elif p.name in ("_warnings.c", "_sre.c", "pyexpat.c", "_bsddb.c"):
                    names.append(p.with_suffix("").name)
                elif p.name in ("socketmodule.c", "posixmodule.c"):
                    names.append(p.name.split("module")[0])
                else:
                    print(f"Unknown module for {s} in {p}, skipped")

    # Some names are listed differently/better here; cjkcodecs and _io/io
    for path in (
        base_path / "PC" / "config.c",
        base_path / "PC" / "os2vacpp" / "config.c",
        base_path / "Python" / "frozen.c",
    ):
        if not path.exists():
            continue
        found = False
        for block in INITTAB_SELECTOR_RE.findall(path.read_text()):
            found = True
            for match in INITTAB_RE.finditer(block):
                if match.group(1) == "__main__":
                    continue
                names.append(match.group(1).split(".")[0])
        if not found:
            print(f"Missing inittab in {path}")

    write_tmpl(f"py{version.replace('.', '')}", version, set(names))
    print(f"{version} done.")
    return set(names)


class ExtensionVisitor(cst.CSTVisitor):
    def __init__(self) -> None:
        self.extension_names: List[str] = []

    def visit_Call(self, node: cst.Call) -> None:
        # print(node)
        d = m.extract(
            node,
            m.Call(
                func=m.OneOf(m.Name("Extension"), m.Name("addMacExtension")),
                args=(
                    m.Arg(value=m.SaveMatchedNode(m.SimpleString(), "extension_name")),
                    m.ZeroOrMore(m.DoNotCare()),
                ),
            ),
        )
        if d:
            assert isinstance(d["extension_name"], cst.SimpleString)
            self.extension_names.append(d["extension_name"].evaluated_value)

    def visit_Assign(self, node: cst.Assign) -> None:
        d = m.extract(
            node,
            m.Assign(
                targets=(m.AssignTarget(target=m.Name("CARBON_EXTS")),),
                value=m.SaveMatchedNode(m.List(), "list"),
            ),
        )
        if d:
            assert isinstance(d["list"], cst.List)
            for item in d["list"].elements:
                if isinstance(item.value, cst.SimpleString):
                    self.extension_names.append(item.value.evaluated_value)


# This is from usort
def try_parse(path: Path, data: Optional[bytes] = None) -> cst.Module:
    """
    Attempts to parse the file with all syntax versions known by LibCST.

    If parsing fails on all supported grammar versions, then raises the parser error
    from the first/newest version attempted.
    """
    if data is None:
        data = path.read_bytes()

    parse_error: Optional[cst.ParserSyntaxError] = None

    for version in cst.KNOWN_PYTHON_VERSION_STRINGS[::-1]:
        try:
            mod = cst.parse_module(
                data, cst.PartialParserConfig(python_version=version)
            )
            return mod
        except cst.ParserSyntaxError as e:
            # keep the first error we see in case parsing fails on all versions
            if parse_error is None:
                parse_error = e

    # not caring about existing traceback here because it's not useful for parse
    # errors, and usort_path is already going to wrap it in a custom class
    raise parse_error or Exception("unknown parse failure")


if __name__ == "__main__":
    regen_all()



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

# Generated by stdlibs/fetch.py

from typing import FrozenSet

module_names: FrozenSet[str] = frozenset(
    [
        "AL",
        "Audio_mac",
        "BaseHTTPServer",
        "Bastion",
        "CD",
        "CDIO",
        "CDROM",
        "CGIHTTPServer",
        "CL",
        "CL_old",
        "Canvas",
        "Carbon",
        "ColorPicker",
        "ConfigParser",
        "Cookie",
        "DEVICE",
        "DLFCN",
        "Dialog",
        "DocXMLRPCServer",
        "ERRNO",
        "EasyDialogs",
        "FILE",
        "FL",
        "FileDialog",
        "FixTk",
        "FrameWork",
        "GET",
        "GL",
        "GLWS",
        "HTMLParser",
        "IN",
        "IOCTL",
        "MacOS",
        "MimeWriter",
        "MiniAEFrame",
        "Nav",
        "OSATerminology",
        "PixMapWrapper",
        "Queue",
        "SOCKET",
        "STROPTS",
        "SUNAUDIODEV",
        "SV",
        "ScrolledText",
        "SimpleDialog",
        "SimpleHTTPServer",
        "SimpleXMLRPCServer",
        "SocketServer",
        "StringIO",
        "TYPES",
        "Tix",
        "Tkconstants",
        "Tkdnd",
        "Tkinter",
        "UserDict",
        "UserList",
        "UserString",
        "WAIT",
        "_AE",
        "_AH",
        "_App",
        "_CF",
        "_CG",
        "_CarbonEvt",
        "_Cm",
        "_Ctl",
        "_Dlg",
        "_Drag",
        "_Evt",
        "_File",
        "_Fm",
        "_Folder",
        "_Help",
        "_IBCarbon",
        "_Icn",
        "_LWPCookieJar",
        "_Launch",
        "_List",
        "_Menu",
        "_Mlte",
        "_MozillaCookieJar",
        "_OSA",
        "_Qd",
        "_Qdoffs",
        "_Qt",
        "_Res",
        "_Scrap",
        "_Snd",
        "_TE",
        "_Win",
        "__builtin__",
        "__future__",
        "__hello__",
        "__phello__",
        "_ast",
        "_bisect",
        "_bsddb",
        "_codecs",
        "_codecs_cn",
        "_codecs_hk",
        "_codecs_iso2022",
        "_codecs_jp",
        "_codecs_kr",
        "_codecs_tw",
        "_csv",
        "_ctypes",
        "_ctypes_test",
        "_curses",
        "_curses_panel",
        "_elementtree",
        "_emx_link",
        "_functools",
        "_hashlib",
        "_heapq",
        "_hotshot",
        "_locale",
        "_lsprof",
        "_md5",
        "_msi",
        "_multibytecodec",
        "_random",
        "_sha",
        "_sha256",
        "_sha512",
        "_socket",
        "_sqlite3",
        "_sre",
        "_ssl",
        "_strptime",
        "_struct",
        "_subprocess",
        "_symtable",
        "_testcapi",
        "_threading_local",
        "_tkinter",
        "_types",
        "_weakref",
        "_winreg",
        "aepack",
        "aetools",
        "aetypes",
        "aifc",
        "al",
        "anydbm",
        "applesingle",
        "appletrawmain",
        "appletrunner",
        "argvemulator",
        "array",
        "asynchat",
        "asyncore",
        "atexit",
        "audiodev",
        "audioop",
        "autoGIL",
        "base64",
        "bdb",
        "bgenlocations",
        "binascii",
        "binhex",
        "bisect",
        "bsddb",
        "bsddb185",
        "buildtools",
        "bundlebuilder",
        "bz2",
        "cPickle",
        "cProfile",
        "cStringIO",
        "calendar",
        "cd",
        "cddb",
        "cdplayer",
        "cfmfile",
        "cgi",
        "cgitb",
        "chunk",
        "cl",
        "cmath",
        "cmd",
        "code",
        "codecs",
        "codeop",
        "collections",
        "colorsys",
        "commands",
        "compileall",
        "compiler",
        "contextlib",
        "cookielib",
        "copy",
        "copy_reg",
        "crypt",
        "csv",
        "ctypes",
        "curses",
        "datetime",
        "dbhash",
        "dbm",
        "decimal",
        "difflib",
        "dircache",
        "dis",
        "distutils",
        "dl",
        "doctest",
        "dumbdbm",
        "dummy_thread",
        "dummy_threading",
        "email",
        "encodings",
        "errno",
        "exceptions",
        "fcntl",
        "filecmp",
        "fileinput",
        "findertools",
        "fl",
        "flp",
        "fm",
        "fnmatch",
        "formatter",
        "fpectl",
        "fpetest",
        "fpformat",
        "ftplib",
        "functools",
        "gc",
        "gdbm",
        "gensuitemodule",
        "gestalt",
        "getopt",
        "getpass",
        "gettext",
        "gl",
        "glob",
        "gopherlib",
        "grp",
        "gzip",
        "hashlib",
        "heapq",
        "hmac",
        "hotshot",
        "htmlentitydefs",
        "htmllib",
        "httplib",
        "ic",
        "icglue",
        "icopen",
        "idlelib",
        "ihooks",
        "imageop",
        "imaplib",
        "imgfile",
        "imghdr",
        "imp",
        "imputil",
        "inspect",
        "itertools",
        "jpeg",
        "keyword",
        "linecache",
        "linuxaudiodev",
        "locale",
        "logging",
        "macerrors",
        "macfs",
        "macostools",
        "macpath",
        "macresource",
        "macurl2path",
        "mailbox",
        "mailcap",
        "markupbase",
        "marshal",
        "math",
        "md5",
        "mhlib",
        "mimetools",
        "mimetypes",
        "mimify",
        "mmap",
        "modulefinder",
        "msilib",
        "msvcrt",
        "multifile",
        "mutex",
        "netrc",
        "new",
        "nis",
        "nntplib",
        "nt",
        "ntpath",
        "nturl2path",
        "opcode",
        "operator",
        "optparse",
        "os",
        "os2",
        "os2emxpath",
        "ossaudiodev",
        "panel",
        "panelparser",
        "parser",
        "pcre",
        "pdb",
        "pickle",
        "pickletools",
        "pimp",
        "pipes",
        "pkgutil",
        "platform",
        "plistlib",
        "popen2",
        "poplib",
        "posix",
        "posixfile",
        "posixpath",
        "pprint",
        "profile",
        "pstats",
        "pty",
        "pure",
        "pwd",
        "py_compile",
        "pyclbr",
        "pydoc",
        "pyexpat",
        "quopri",
        "random",
        "re",
        "readcd",
        "readline",
        "repr",
        "resource",
        "rexec",
        "rfc822",
        "rgbimg",
        "riscosenviron",
        "riscospath",
        "rlcompleter",
        "robotparser",
        "rourl2path",
        "runpy",
        "sched",
        "select",
        "sets",
        "sgi",
        "sgmllib",
        "sha",
        "shelve",
        "shlex",
        "shutil",
        "signal",
        "site",
        "smtpd",
        "smtplib",
        "sndhdr",
        "socket",
        "spwd",
        "sqlite3",
        "sre",
        "sre_compile",
        "sre_constants",
        "sre_parse",
        "stat",
        "statvfs",
        "string",
        "stringold",
        "stringprep",
        "strop",
        "struct",
        "subprocess",
        "sunau",
        "sunaudio",
        "sunaudiodev",
        "sv",
        "symbol",
        "symtable",
        "sys",
        "syslog",
        "tabnanny",
        "tarfile",
        "telnetlib",
        "tempfile",
        "terminalcommand",
        "termios",
        "textwrap",
        "this",
        "thread",
        "threading",
        "time",
        "timeit",
        "timing",
        "tkColorChooser",
        "tkCommonDialog",
        "tkFileDialog",
        "tkFont",
        "tkMessageBox",
        "tkSimpleDialog",
        "toaiff",
        "token",
        "tokenize",
        "torgb",
        "trace",
        "traceback",
        "tty",
        "turtle",
        "types",
        "unicodedata",
        "unittest",
        "urllib",
        "urllib2",
        "urlparse",
        "user",
        "uu",
        "uuid",
        "videoreader",
        "warnings",
        "wave",
        "weakref",
        "webbrowser",
        "whichdb",
        "winsound",
        "wsgiref",
        "xdrlib",
        "xml",
        "xmllib",
        "xmlrpclib",
        "xx",
        "xxsubtype",
        "zipfile",
        "zipimport",
        "zlib",
    ]
)
"""
Known stdlib modules for Python 2.5.
"""



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

# Generated by stdlibs/fetch.py

from typing import List

KNOWN_VERSIONS: List[str] = [
    "2.3",
    "2.4",
    "2.5",
    "2.6",
    "2.7",
    "3.0",
    "3.1",
    "3.2",
    "3.3",
    "3.4",
    "3.5",
    "3.6",
    "3.7",
    "3.8",
    "3.9",
    "3.10",
    "3.11",
]
"""All supported Python major releases"""



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

"""
List of packages in the stdlib
"""

__author__ = "Amethyst Reese"

import importlib
import sys
from typing import FrozenSet, List, Optional

from . import known

from .__version__ import __version__

ALL = "all"
"""Combined 3.x and 2.x stdlibs"""

KNOWN_VERSIONS: List[str] = known.KNOWN_VERSIONS
"""All supported Python major releases"""


def stdlib_module_names(version: Optional[str] = None) -> FrozenSet[str]:
    """
    Return a set of known module names for a Python release in :data:`.KNOWN_VERSIONS`.

    If passed :data:`stdlibs.ALL`, this returns :data:`stdlibs.py.module_names`.
    If passed ``None``, or no arguments, returns the equivalent module names for the
    active major Python version.

    Prefer using :data:`stdlibs.module_names` or a specific
    :ref:`Versioned Module <versioned>` directly.
    """
    if version is None:
        version = "%d%d" % sys.version_info[:2]
        modname = f".py{version}"
    elif version == ALL:
        modname = ".py"
    else:
        version = "".join(version.split(".")[:2])
        modname = f".py{version}"

    return importlib.import_module(modname, __package__).module_names  # type: ignore


module_names: FrozenSet[str] = stdlib_module_names("3")
"""
Known stdlibs for any release of Python 3.x

Shortcut for :data:`stdlibs.py3.module_names`.
"""


__all__ = [
    "ALL",
    "KNOWN_VERSIONS",
    "module_names",
    "stdlib_module_names",
]



# Copyright 2022 Amethyst Reese
# Licensed under the MIT license

import os
import sys
from pathlib import Path
from unittest import mock, TestCase

import stdlibs


class StdlibsTest(TestCase):
    def test_module_names(self) -> None:
        for name in ("re", "os", "sys", "posixpath", "re", "_sre", "xmlrpc"):
            self.assertIn(name, stdlibs.module_names)
        self.assertNotIn("__main__", stdlibs.module_names)

    def test_default_version(self) -> None:
        with mock.patch("stdlibs.sys.version_info", (3, 7, 5)):
            names = stdlibs.stdlib_module_names()
        self.assertEqual(names, stdlibs.stdlib_module_names("3.7"))

    def test_api(self) -> None:
        names = stdlibs.stdlib_module_names("3.8")
        self.assertIn("posixpath", names)

        # Error handling
        with self.assertRaises(ImportError):
            stdlibs.stdlib_module_names("1.0")

    def test_all_my_modules_are_there(self) -> None:
        # Gets only the top-level names we should have on this version
        names = stdlibs.stdlib_module_names(None)
        for module in Path(os.__file__).parent.glob("*.py"):
            name = module.with_suffix("").name
            name = name.split(".")[0]  # __phello__.foo
            if name.startswith("_sysconfigdata_"):
                continue
            self.assertIn(name, names)

    def test_all23(self) -> None:
        self.assertEqual(
            stdlibs.stdlib_module_names("2") | stdlibs.stdlib_module_names("3"),
            stdlibs.stdlib_module_names("all"),
        )

    def test_readme_example(self) -> None:
        self.assertEqual(
            ["3.7", "3.8", "3.9", "3.10", "3.11"],
            [
                v
                for v in stdlibs.KNOWN_VERSIONS
                if "dataclasses" in stdlibs.stdlib_module_names(v)
            ],
        )
